(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                 A Functional Constraint Library                     *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, CENA               *)
(*                                                                     *)
(* Copyright 2001 CENA. All rights reserved. This file is distributed  *)
(* under the terms of the GNU Lesser General Public License.           *)
(***********************************************************************)
(* $Id: fcl_arith.ml,v 1.53 2001/06/15 10:22:31 brisset Exp $ *)

open Fcl_var
open Fcl_misc.Operators
open Printf

type t =
    Int of int
  | Fd of Fd.t
  | Plus of t * t
  | Mult of t * t
  | Minus of t
  | Div of t * t
  | Abs of t
  | Mod of t * t

type t' = t

module T_for_H = struct
  module I = struct
    type ti = (* Replacing Fd.t by int to be able to hash *)
      	Int of int
      | Fd of int
      | Plus of ti * ti
      | Mult of ti * ti
      | Minus of ti
      | Div of ti * ti
      | Abs of ti
      |	Mod of ti * ti
  end
  let rec i = function
      Int int -> I.Int int
    | Fd v -> I.Fd (Fd.id v)
    | Plus (e1,e2) -> I.Plus (i e1,i e2)
    | Mult (e1,e2) -> I.Mult (i e1,i e2)
    | Minus e -> I.Minus (i e)
    | Div (e1,e2) -> I.Div (i e1,i e2)
    | Abs e -> I.Abs (i e)
    | Mod (e1,e2) -> I.Mod (i e1,i e2)

  type t = t'

  let hash e =
    Hashtbl.hash (i e)

  let rec equal e1 e2 =
    match (e1, e2) with
      (Int i1, Int i2) -> i1 = i2
    | (Fd v1, Fd v2) -> Fd.id v1 = Fd.id v2
    | Plus (e11, e12), Plus (e21, e22)-> equal e11 e21 && equal e12 e22
    | Mult (e11, e12), Mult (e21, e22)-> equal e11 e21 && equal e12 e22
    | Minus e1, Minus e2 -> equal e1 e2
    | Div (e11, e12), Div (e21, e22)-> equal e11 e21 && equal e12 e22
    | Abs e1, Abs e2 -> equal e1 e2
    | Mod (e11, e12), Mod (e21, e22)-> equal e11 e21 && equal e12 e22
    | _ -> false
end

module H_of_t = Hashtbl.Make(T_for_H)

let lowest_priority = 0
let priority = function
  | Plus (_e1, _e2) -> 1
  | Mult (_, _) | Div (_, _) | Mod (_, _) -> 2
  | Minus _e -> 3
  | _ -> 4
type associativity = Left | Right | None | Both
let associativity = function
  | Plus (_, _) | Mult (_, _) -> Both
  | Div (_, _) | Mod (_, _) -> Left
  | Minus _e -> Right
  | _ -> None



type right_hand =
    Division of Fd.t * Fd.t
  | Monome of Fd.t * Fd.t
  | Expn of Fd.t * int
  | Absolute of Fd.t
  | Modulo of Fd.t * Fd.t

type equation =
    Fd.t * right_hand

let to_be_posted = ref ([] : equation list)
let add_to_be_posted c =
  to_be_posted := c :: !to_be_posted


let left p = function
    Left | Both -> p
  | Right -> p+1
  | None -> lowest_priority (* unused *)

let right p = function
    Right | Both -> p
  | Left -> p+1
  | None -> lowest_priority (* unused *)


let fprint c e =
  let rec loop prio_father e =
    let prio = priority e
    and assoc = associativity e in
    let needs_parenthesis = prio_father > prio
    and loop_left = fun _channel -> loop (left prio assoc)
    and loop_right = fun _channel -> loop (right prio assoc) in
    if needs_parenthesis then Printf.fprintf c "(";
    begin
      match e with
      	Int x -> Printf.fprintf c "%d" x
      | Fd fd -> Fd.fprint c fd
      | Plus (e1, e2) -> Printf.fprintf c "%a+%a" loop_left e1 loop_right e2
      | Mult (e1, e2) -> Printf.fprintf c "%a*%a" loop_left e1 loop_right e2
      | Minus e -> Printf.fprintf c "-%a" loop_right e
      | Div (e1, e2) -> Printf.fprintf c "%a/%a" loop_left e1 loop_right e2
      | Abs e -> Printf.fprintf c "|%a|" loop_right e
      | Mod (e1, e2) -> Printf.fprintf c "%a%%%a" loop_left e1 loop_right e2
    end;
    if needs_parenthesis then Printf.fprintf c ")" in
  loop lowest_priority e

let sign x = if x < 0 then (-1) else if x = 0 then 0 else 1

let rec eval = function
    Int x -> x
  | Fd fd -> (match Fd.value fd with Unk _ -> raise (Invalid_argument "Arith.eval: not ground") | Val v -> v)
  | Plus (e1, e2) -> eval e1 + eval e2
  | Mult (e1, e2) -> eval e1 * eval e2
  | Minus e -> 0 - eval e
  | Div (e1, e2) -> eval e1 / eval e2
  | Abs e -> abs (eval e)
  | Mod (e1, e2) -> eval e1 mod eval e2


let rec min_of_expr = function
    Int x -> x
  | Fd x -> (Fd.min x)
  | Plus (x1,x2) -> min_of_expr x1 + min_of_expr x2
  | Minus e -> 0 - (max_of_expr e)
  | Mult (x1, x2) ->
      let a = min_of_expr x1 and b = max_of_expr x1
      and c = min_of_expr x2 and d = max_of_expr x2 in
      if sign a * sign b <= 0 || sign c * sign d <= 0 then
      (* L'un des deux domaines contient 0 *)
        min (a*d) (b*c)
      else if sign a * sign c < 0 then
      (* Les deux domaines ne sont pas du meme cote par rapport a 0 *)
        if a > 0 then b*c else a*d
      else
        if a > 0 then a*c else b*d
  | Div (x1, x2) ->
      let a = min_of_expr x1 and b = max_of_expr x1
      and c = min_of_expr x2 and d = max_of_expr x2 in
      if c = d && c = 0 then raise Division_by_zero else
      (* Sinon on suppose que x2 ne sera pas instanci  0 *)
      let c = if c = 0 then 1 else c and d = if d = 0 then -1 else d in
      if sign c * sign d < 0 then
	min a (0 - b)
      else if sign a * sign b < 0 then
	if c > 0 then a/c else b/d
      else
	if a >= 0 then (* x1 positif *)
	  if c > 0 then a/d else b/d (* x2 positif ou ngatif *)
	else (* x1 ngatif *)
	  if c < 0 then b/c else a/c (* x2 ngatif ou positif *)
  | Abs x ->
      let a = min_of_expr x in
      if a >= 0 then a
      else
	let b = max_of_expr x in
	if b <= 0 then -b else 0
  | Mod (x1, x2) ->
      let abs_x2 = Abs x2 and a = min_of_expr x1 in
      if a >= 0 then (* x1 >= 0 *)
	min_of_absmod x1 abs_x2
      else
	let b = max_of_expr x1 in
	if b <= 0 then (* x1 <= 0 *)
	  let abs_x1 = Abs x1 in
	  0 - (max_of_absmod abs_x1 abs_x2)
	else
	  let negx1 = Fd (Fd.interval 0 (-a)) in
	  0 - (max_of_absmod negx1 abs_x2)
and max_of_expr = function
    Int x -> x
  | Fd x -> (Fd.max x)
  | Plus (x1,x2) -> max_of_expr x1 + max_of_expr x2
  | Minus e -> 0 - (min_of_expr e)
  | Mult (x1, x2) ->
      let a = min_of_expr x1 and b = max_of_expr x1
      and c = min_of_expr x2 and d = max_of_expr x2 in
      if sign a * sign b <= 0 || sign c * sign d <= 0 then
      (* L'un des deux domaines est  cheval sur 0 *)
        max (a*c) (b*d)
      else if sign a * sign c < 0 then
      (* Les deux domaines ne sont pas du meme cote par rapport a 0 *)
	if a > 0 then a*d else b*c
      (* Les deux domaines sont du meme cote *)
      else
        if a > 0 then b*d else a*c
  | Div (x1, x2) ->
      let a = min_of_expr x1 and b = max_of_expr x1
      and c = min_of_expr x2 and d = max_of_expr x2 in
      if c = d && c = 0 then raise Division_by_zero else
      (* Sinon on suppose que x2 ne sera pas instanci 0 *)
      let c = if c = 0 then 1 else c and d = if d = 0 then -1 else d in
      if sign c * sign d < 0 then
	max (0 - a) b
      else if sign a * sign b < 0 then
	if c > 0 then b/c else a/d
      else
	if a >= 0 then (* x1 positif *)
	  if c > 0 then b/c else a/c (* x2 positif ou ngatif *)
	else (* x1 ngatif *)
	  if c < 0 then a/d else b/d (* x2 ngatif ou positif *)
  | Abs x ->
      let a = min_of_expr x and b = max_of_expr x in
      if a >= 0 then b else if b <= 0 then (0 - a) else max (0 - a) b
  | Mod (x1, x2) ->
      let abs_x2 = Abs x2 and a = min_of_expr x1 in
      if a >= 0 then (* x1 >= 0 *)
	max_of_absmod x1 abs_x2
      else
	let b = max_of_expr x1 in
	if b <= 0 then (* x1 <= 0 *)
	  let abs_x1 = Abs x1 in
	  0 - (min_of_absmod abs_x1 abs_x2)
	else
	  let posx1 = Fd (Fd.interval 0 b) in
	  max_of_absmod posx1 abs_x2

(* min et max spciaux pour mod quand x1 >= 0 et x2 > 0
   pour une plateforme avec x mod y = x - (x/y)*y comme Linux :
          x mod y = sign x * (abs x) mod (abs y) *)
and min_of_absmod x1 x2 =
  let b = max_of_expr x1 and c = min_of_expr x2 in
  if b < c then min_of_expr x1
  else (* 0 *)
  (* on peut faire mieux quand c=d (i.e. x2 connu) *)
    let d = max_of_expr x2 in
    if c = d then
      if c = 0 then raise Division_by_zero else
      let a = min_of_expr x1 in
      let amodc = a mod c in
      if b - a + amodc < c then amodc else 0
    else 0

and max_of_absmod x1 x2 =
  let b = max_of_expr x1 and d = max_of_expr x2 in
  if b < d then b
  else (* d - 1 *)
  (* on peut faire mieux quand c=d (i.e. x2 connu) *)
    let c = min_of_expr x2 in
    if c = d then
      if c = 0 then raise Division_by_zero else
      let a = min_of_expr x1 and bmodc = b mod c in
      if bmodc - (b - a) > 0 then bmodc else d - 1
    else d - 1


let rec min_max_of_expr = function
    Int x -> (x, x)
  | Fd x -> Fd.min_max x
  | Plus (x1,x2) ->
      let (min_x1, max_x1) = min_max_of_expr x1
      and (min_x2, max_x2) = min_max_of_expr x2 in
      (min_x1 + min_x2, max_x1 + max_x2)
  | Minus e ->
      let (min_e, max_e) = min_max_of_expr e in
      (0 - max_e, 0 - min_e)
  | Mult (x1, x2) ->
      let (a, b) = min_max_of_expr x1
      and (c, d) = min_max_of_expr x2 in
      if sign a * sign b <= 0 || sign c * sign d <= 0 then
      (* L'un des deux domaines contient 0 *)
        (min (a*d) (b*c), max (a*c) (b*d))
      else if sign a * sign c < 0 then
      (* Les deux domaines ne sont pas du meme cote par rapport a 0 *)
        if a > 0 then (b*c, a*d) else (a*d, b*c)
      else
        if a > 0 then (a*c, b*d) else (b*d, a*c)
  | Div (x1, x2) ->
      let (a, b) = min_max_of_expr x1
      and (c, d) = min_max_of_expr x2 in
      if c = d && c = 0 then raise Division_by_zero else
      (* Sinon on suppose que x2 ne sera pas instanci  0 *)
      let c = if c = 0 then 1 else c and d = if d = 0 then -1 else d in
      if sign c * sign d < 0 then
	(min a (0 - b), max (0 - a) b)
      else if sign a * sign b < 0 then
	if c > 0 then (a/c, b/c) else (b/d, a/d)
      else
	if a >= 0 then (* x1 positif *)
	  if c > 0 then (a/d, b/c) else (b/d, a/c) (* x2 positif ou ngatif *)
	else (* x1 ngatif *)
	  if c < 0 then (b/c, a/d) else (a/c, b/d) (* x2 ngatif ou positif *)
  | Abs x ->
      let (a, b) = min_max_of_expr x in
      if a >= 0 then (a, b) else
      if b <= 0 then (0 - b, 0 - a) else (0, max (0 - a) b)
  | Mod (x1, x2) ->
      let abs_x2 = Abs x2 and (a, b) = min_max_of_expr x1 in
      let (c, d) = min_max_of_expr abs_x2 in
      if a >= 0 then (* x1 >= 0 *)
	(min_of_absmod_inter a b c d, max_of_absmod_inter a b c d)
      else
	if b <= 0 then (* x1 <= 0 *)
	  let abs_x1 = Abs x1 in
	  let (a, b) = min_max_of_expr abs_x1 in
	  (0 - max_of_absmod_inter a b c d, 0 - min_of_absmod_inter a b c d)
	else
	  let negx1 = Fd (Fd.interval 0 (-a))
	  and posx1 = Fd (Fd.interval 0 b) in
	  (0 - max_of_absmod_inter 0 (-a) c d, max_of_absmod_inter 0 b c d)

and min_of_absmod_inter a b c d =
  if b < c then a
  else (* 0 *)
  (* on peut faire mieux quand c=d (i.e. x2 connu) *)
    if c = d then
      if c = 0 then raise Division_by_zero else
      let amodc = a mod c in
      if b - a + amodc < c then amodc else 0
    else 0

and max_of_absmod_inter a b c d =
  if b < d then b
  else (* d - 1 *)
  (* on peut faire mieux quand c=d (i.e. x2 connu) *)
    if c = d then
      if c = 0 then raise Division_by_zero else
      let bmodc = b mod c in
      if bmodc - (b - a) > 0 then bmodc else d - 1
    else d - 1

type linear_expression = int * (int*Fd.t) list

type cstr =
    Leq0 of linear_expression
  | Neq0 of linear_expression
  | Eq0 of linear_expression


let (/+) x y =
  let xy = x / y in
  if x mod y = 0 then xy else
  if sign x * sign y >= 0 then xy + 1 else xy

let (/-) x y =
  let xy = x / y in
  if x mod y = 0 then xy else
  if sign x * sign y >= 0 then xy else xy - 1

let div_sup = (/+)

type operator = LessThan | Equal | Diff

let string_of_op op =
  match op with Equal -> "=" | LessThan -> "<=" | Diff -> "<>"

let refine_min_max z za z_min z_max =
  Fcl_debug.call 'a' (fun s -> fprintf s "refine_min_max: %d %d\n" z_min z_max);
  Fd.refine z
    (Fcl_domain.remove_up z_max
       (Fcl_domain.remove_low z_min (Attr.dom za)));;

let refine_up x x_max =
  match Fd.value x with
    Val x -> if x > x_max then Fcl_stak.fail "refine_up"
  | Unk a -> Fd.refine x (Fcl_domain.remove_up x_max (Attr.dom a))

let refine_low x x_min =
  match Fd.value x with
    Val x -> if x < x_min then Fcl_stak.fail "refine_low"
  | Unk a -> Fd.refine x (Fcl_domain.remove_low x_min (Attr.dom a))

let remove_constants terms =
  let modif = ref false in
  let r =
    List.fold_left
      (fun (cst, vars) ((a, x) as ax) ->
	match Fd.value x with
	  Unk _ -> (cst, ax::vars)
	| Val v -> modif := true; (cst+a*v, vars))
      (0, [])
      terms in
  if !modif then r else raise Not_found;;

let compute_inf_sup pos_terms neg_terms =
  let neg_inf_sum,neg_sup_sum = List.fold_left
      (fun (inf,sup) (c,x) ->
	let mi,ma = Fd.min_max x in
	(inf+c*ma, sup+c*mi))
      (0, 0)
      neg_terms in
  List.fold_left
    (fun (inf,sup) (c,x) ->
      let mi,ma = Fd.min_max x in
      (inf+c*mi, sup+c*ma))
    (neg_inf_sum, neg_sup_sum)
    pos_terms;;

let part_pos_neg l =
  let rec loop pos neg = function
      [] -> (pos, neg)
    | (0, _) :: axs -> loop pos neg axs
    | (a, _x) as ax :: axs ->
	if a > 0 then loop (ax :: pos) neg axs
	else loop pos (ax :: neg) axs in
  loop [] [] l

(* In case all terms are positive (resp. negative), one can try to refine
   individually each term (avoids an integer overflow in magic sequence caused
   by a giant and useless scalar product) *)
let basic_refinements pos_terms neg_terms d op =
  if op <> Diff then
    if List.for_all (fun (_a, x) -> Fd.min x >= 0 ) pos_terms &&
      List.for_all (fun (_a, x) -> Fd.max x <= 0 ) neg_terms then begin
	List.iter (fun (a, x) -> refine_up x (d/a)) pos_terms;
	List.iter (fun (a, x) -> refine_low x (d/a)) neg_terms
      end
    else if op = Equal && (* so GreaterThan *)
      List.for_all (fun (_a, x) -> Fd.max x <= 0 ) pos_terms &&
      List.for_all (fun (_a, x) -> Fd.min x >= 0 ) neg_terms then begin
	List.iter (fun (a, x) -> refine_low x (d/a)) pos_terms;
	List.iter (fun (a, x) -> refine_up x (d/a)) neg_terms
      end


class linear (terms : (int*Fd.t) list) op dd =
  let pos_terms, neg_terms = part_pos_neg terms in
  let pos_terms = Fcl_stak.ref pos_terms
  and neg_terms = Fcl_stak.ref neg_terms
  and d = Fcl_stak.ref dd in

  object
  inherit Fcl_cstr.t Fcl_cstr.normal
  method name = "linear"
  method delay c =
    List.iter
      (fun (a, x) ->
	match op with
	  Diff -> delay [Attr.on_subst] x c
	| Equal -> delay [Attr.on_refine] x c
	| LessThan ->
	    delay [if a > 0 then Attr.on_min else Attr.on_max] x c)
      terms
  method fprint c =
    List.iter (fun (a,x) -> Printf.fprintf c " +%d.%a" a Fd.fprint x) (Fcl_stak.get pos_terms);
    List.iter (fun (a,x) -> Printf.fprintf c " %d.%a" a Fd.fprint x) (Fcl_stak.get neg_terms);
    Printf.fprintf c " %s %d" (string_of_op op) (Fcl_stak.get d);
    flush c
  method init =
    Fcl_debug.call 'a' (fun s -> fprintf s "linear#init: %a %s %d\n" (fun s -> List.iter (fun (c, v) -> fprintf s "%d.%a " c Fd.fprint v)) terms (match op with Diff -> "<>" | Equal -> "=" | LessThan -> "<=") dd);
    match terms with
      [] -> begin
	Fcl_stak.set solved true;
      	match op with
      	  Equal -> if dd <> 0 then Fcl_stak.fail "linear Equal d<>0"
      	| Diff -> if dd = 0 then Fcl_stak.fail "linear Diff d = 0"
      	| LessThan -> if 0 > dd then Fcl_stak.fail "linear LessThan 0 > d"
      end
    | _terms -> basic_refinements (Fcl_stak.get pos_terms) (Fcl_stak.get neg_terms) (Fcl_stak.get d) op

  method not =
    let terms = Fcl_stak.get pos_terms @ Fcl_stak.get neg_terms
    and d = Fcl_stak.get d in
    match op with
      Equal -> new linear terms Diff d
    | Diff -> new linear terms Equal d
    | LessThan -> new linear (List.map (fun (a,x) -> (-a, x)) terms) LessThan (-1 - d)

  method check =
    let inf_sum, sup_sum = compute_inf_sup (Fcl_stak.get pos_terms) (Fcl_stak.get neg_terms) in
    let d = Fcl_stak.get d in
    match op with
      Equal ->
	if inf_sum = sup_sum then inf_sum = d else
	if sup_sum < d || inf_sum > d then false else
	raise Fcl_cstr.DontKnow
    | Diff ->
 	if inf_sum = sup_sum then inf_sum <> d else
	sup_sum < d || inf_sum > d || raise Fcl_cstr.DontKnow
    | LessThan ->
	sup_sum <= d || (if inf_sum > d then false else (Fcl_debug.call 'a' (fun s -> fprintf s "Dont know\n"); raise Fcl_cstr.DontKnow))

  method update =
    begin
      try
	let (cst, new_pos_terms) = remove_constants (Fcl_stak.get pos_terms) in
	if cst <> 0 then Fcl_stak.set d (Fcl_stak.get d - cst);
	Fcl_stak.set pos_terms new_pos_terms
      with Not_found -> () (* No constant positive term *)
    end;
    begin
      try
	let (cst, new_neg_terms) = remove_constants (Fcl_stak.get neg_terms) in
	if cst <> 0 then Fcl_stak.set d (Fcl_stak.get d - cst);
	Fcl_stak.set neg_terms new_neg_terms
      with Not_found -> () (* No constant negative term *)
    end;

    let d = Fcl_stak.get d in
    match Fcl_stak.get pos_terms, Fcl_stak.get neg_terms with
      ([], []) ->
	(match op with
	  Diff when d <> 0 -> Fcl_stak.set solved true
	| Equal when d = 0 -> Fcl_stak.set solved true
	| LessThan when d >= 0 -> Fcl_stak.set solved true
	| _ -> Fcl_stak.fail "linear[]")
    | ([(a, x)],[]) when op = Equal ->
	if d mod a = 0
	then (Fcl_stak.set solved true; Fd.subst x (d/a))
	else Fcl_stak.fail "linear[(a,x)]"
    | ([(a, x)],[]) when op = Diff ->
	if d mod a = 0
	then (
	  match Fd.value x with
	    Val _ -> Fcl_debug.internal_error "linear#update ([(a, x)],[]) when op = Diff"
	  | Unk attr -> Fcl_stak.set solved true; Fd.refine x (Fcl_domain.remove (d/a) (Attr.dom attr))
		)
	else Fcl_stak.set solved true
    | ([],[(a, x)]) when op = Equal ->
	if d mod (-a) = 0
	then (Fcl_stak.set solved true; Fd.subst x (d/a))
	else Fcl_stak.fail "linear[(a,x)]"
    | ([(1, x1)], [(-1,x2)]) when op = Equal -> (* x1 = x2 + d *)
	(*** Printf.printf "x1 = x2 + d"; ***)
	(match Fd.value x1, Fd.value x2 with
	  Unk a1, Unk a2 ->
	    let d1 = Attr.dom a1 and d2 = Attr.dom a2 in
	    Fd.refine x1 (Fcl_domain.intersection d1 (Fcl_domain.plus d2 d));
	    Fd.refine x2 (Fcl_domain.intersection d2 (Fcl_domain.plus d1 (-d)))
	| _ -> Fcl_debug.internal_error "Arith 1 -1")
    | ([(1, x1)], [(-1,x2)]) when op = LessThan -> (* x1 <= x2 + d *)
	(match Fd.value x1, Fd.value x2 with
	  Unk a1, Unk a2 ->
	    let max1 = Attr.max a1 and min2 = Attr.min a2
	    and max2 = Attr.max a2 and min1 = Attr.min a1 in
	    if max1 > max2 + d then
	      Fd.refine x1 (Fcl_domain.remove_up (max2+d) (Attr.dom a1));
	    if min1 > min2 + d then
	      Fd.refine x2 (Fcl_domain.remove_low (min1-d) (Attr.dom a2));
	    if Attr.max a1 <= Attr.min a2 + d then
	      Fcl_stak.set solved true
	| _ -> Fcl_debug.internal_error "Arith 1 -1")
    | ([], [(a, x)]) when op = Diff ->
	if d mod (-a) = 0
	then (
	  match Fd.value x with
	    Val _ -> Fcl_debug.internal_error "linear#update ([], [(a, x)]) when op = Diff"
	  | Unk attr -> Fcl_stak.set solved true; Fd.refine x (Fcl_domain.remove (d/a) (Attr.dom attr))
		)
	else Fcl_stak.set solved true
    | _ ->	  
	
	if op = Diff then (* on attend instantiation *)
	  ()
	else begin
	  let modif = ref true (* Only for Equal *)
	  and instantiated = ref true in
	  while !modif || !instantiated do
	    let last_modif = !modif in
	    modif := false;
	    instantiated := false;
	    let inf_sum, sup_sum = compute_inf_sup (Fcl_stak.get pos_terms) (Fcl_stak.get neg_terms) in    
	    let d_inf_sum = d - inf_sum and d_sup_sum = d - sup_sum in
	    if op = LessThan && 0 <= d_sup_sum || d_sup_sum = 0 && d_inf_sum = 0 then
		Fcl_stak.set solved true
	    else if 0 > d_inf_sum || op = Equal && 0 < d_sup_sum then
	      Fcl_stak.fail "linear d_inf_sum"
	    else if last_modif then begin (* We stop here if only instantiated *)
	      (* Let's update bounds of variables *)
	      let update_pos (a, x) =
		match Fd.value x with
		    Unk ax ->
		      let domx = Attr.dom ax in
		      let mi = Fcl_domain.min domx and ma = Fcl_domain.max domx in
		      let new_sup = min ma ((d_inf_sum + a*mi)/a) in
		      if op = Equal then begin
			let new_inf = max mi (div_sup (d_sup_sum + a*ma) a) in
			if new_sup < ma || new_inf > mi then begin
			  modif:=true;
			  refine_min_max x ax new_inf new_sup
			end
		      end else begin
			if new_sup < ma then begin
			  Fd.refine x (Fcl_domain.remove_up new_sup domx);
			  if not (Fd.is_var x) then instantiated := true
 (* because it's maybe the last variable of the expression and the constraint
    is now solved *)	
			end
		      end
		  | Val vx ->
		      let new_sup = ((d_inf_sum + a*vx)/a) in
		      if new_sup < vx then Fcl_stak.fail "Arith.update_pos sup";
		      if op = Equal then
			let new_inf = (div_sup (d_sup_sum + a*vx) a) in
			if new_inf > vx then Fcl_stak.fail "Arith.update_pos inf" in
	      List.iter update_pos (Fcl_stak.get pos_terms);
	      
	      let update_neg (a, x) =
		match Fd.value x with
		    Unk ax ->
		      let domx = Attr.dom ax in
		      let mi = Fcl_domain.min domx and ma = Fcl_domain.max domx in
		      let new_inf = max mi (div_sup (d_inf_sum + a*ma) a) in
		      if op = Equal then begin
			let new_sup = min ma ((d_sup_sum + a*mi)/a) in
			if new_sup < ma || new_inf > mi then begin
			  modif:=true;
			  refine_min_max x ax new_inf new_sup
			end
		      end else begin
			if new_inf > mi then begin
			  Fcl_debug.call 'a' (fun s -> fprintf s "linear#update, refine_low %d\n" new_inf);
			  Fd.refine x (Fcl_domain.remove_low new_inf domx);
			  if not (Fd.is_var x) then instantiated := true
 (* because it's maybe the last variable of the expression and the constraint
    is now solved *)	
			end
		      end
		  | Val vx ->
		      let new_inf = (div_sup (d_inf_sum + a*vx) a) in
		      if new_inf > vx then Fcl_stak.fail "Arith.update_neg inf";
		      if op = Equal then
			let new_sup = ((d_sup_sum + a*vx)/a) in
			if new_sup < vx then Fcl_stak.fail "Arith.update_neg sup" in
	      List.iter update_neg (Fcl_stak.get neg_terms)
	    end
	  done
	end
end;;

(* y = z[a,b] / x[c,d] *)
(* min et max pour le refine de y *)
let min_max_of_div_for_mult z x =
  (* Si 0 appartient au domaine de x, on ne peut rien dire sur y *)
  if Fd.member x 0 then
    (min_int, max_int)
  else
    let (a, b) = Fd.min_max z and (c, d) = Fd.min_max x in
    (* Est-ce utile ? On ne rveille que sur min ou max *)
    (* x  cheval sur 0 *)
    if sign c * sign d < 0 then
      (* si x est instanci, c*d est positif *)
      match Fd.value x with
	Unk xa ->
	  let domx = Attr.dom xa in
	  let c' = Fcl_domain.greatest_leq domx 0
	  and d' = Fcl_domain.smallest_geq domx 0 in
	  (* x ne contient pas 0 *)
	  (min (a /+ d') (b /+ c'), max (a /- c') (b /- d'))
      |	_ -> Fcl_debug.internal_error "min_max_of_div_for_mult : x ground"
    (* x positif ou ngatif *)
    (* z  cheval sur 0 *)
    else if sign a * sign b < 0 then
      if c > 0 then (a /+ c, b /- c) else (b /+ d, a /- d)
    (* ni z ni x  cheval sur 0 *)
    else
      if a >= 0 then (* z positif *)
	if c > 0 then (a /+ d, b /- c) else (b /+ d, a /- c) (* x positif ou ngatif *)
      else (* z ngatif *)
	if c < 0 then (b /+ c, a /- d) else (a /+ c, b /- d) (* x ngatif ou positif *)


let unifine z z_min z_max fail_msg =
  match Fd.value z with
    Val c -> if c > z_max || c < z_min then Fcl_stak.fail fail_msg
  | Unk za -> refine_min_max z za z_min z_max;;


(* z = x*y *)
class monome z x y =

  let mult = Mult (Fd x, Fd y)

  and zero_removed = Fcl_stak.ref false

  and update_val_unk a y solved =
    match Fd.value z with
      Val c -> (* a <> 0 *)
	if c mod a = 0 then (Fcl_stak.set solved true; Fd.unify y (c / a))
	else Fcl_stak.fail "monome : Val a, Unk _, Val c"
    | Unk _ ->
	(Fcl_stak.set solved true;
	 Fcl_cstr.post (new linear [(1, z); (-a, y)] Equal 0))

  and compute_bounds () =
    (Fd.min z, Fd.max z, Fd.min x, Fd.max x, Fd.min y, Fd.max y) in

  object
  inherit Fcl_cstr.t Fcl_cstr.normal

  method name = "monome"

  method delay c =
    delay [Attr.on_min] z c; delay [Attr.on_max] z c;
    delay [Attr.on_min] x c; delay [Attr.on_max] x c;
    delay [Attr.on_min] y c; delay [Attr.on_max] y c

  method fprint c =
    Printf.fprintf c "%a = %a * %a" Fd.fprint z Fd.fprint x Fd.fprint y;
    flush c

  method update =
    Fcl_debug.call 'a' (fun s -> fprintf s "monome - before update : %a = %a * %a\n" Fd.fprint z Fd.fprint x Fd.fprint y);

    let rec loop () =
      match Fd.value x, Fd.value y with
	Val a, Val b -> Fcl_stak.set solved true; Fd.unify z (a * b)
      | (Val 0, _  | _, Val 0) -> Fcl_stak.set solved true; Fd.unify z 0
      | Val a, Unk _ -> update_val_unk a y solved
      | Unk _, Val b -> update_val_unk b x solved
(* On pourrait aussi traiter le cas z = Val 1 (idem pour expn) *)
      | Unk xa, Unk ya ->
	  match Fd.value z with
	    Val 0 ->
	      let xa_with_0 = Attr.member xa 0
	      and ya_with_0 = Attr.member ya 0 in
	      if xa_with_0 then begin
		if not ya_with_0 then begin
		  Fcl_stak.set solved true; Fd.unify x 0 end end
	      else if ya_with_0 then begin
		Fcl_stak.set solved true; Fd.subst y 0 end
	      else Fcl_stak.fail "monome : Unk xa, Unk ya, Val c"
	  | _z_val ->
	      (* On essaie d'enlever d'abord 0 dans x et y sinon pas de
		 propagation *)
	      if not (Fcl_stak.get zero_removed) && not (Fd.member z 0) then begin
		Fcl_stak.set zero_removed true;
		Fd.refine x (Fcl_domain.remove 0 (Attr.dom xa));
		Fd.refine y (Fcl_domain.remove 0 (Attr.dom ya));
	        (* x et y peuvent etre instancies *)
		loop () end
	      else begin
		let bounds = compute_bounds () in
		let (z_min, z_max) = min_max_of_expr mult in
		Fcl_debug.call 'a' (fun s -> fprintf s "monome - Unk xa, Unk ya, z : z_min=%d z_max=%d " z_min z_max);
		unifine z z_min z_max "monome : Unk xa, Unk ya, Val c";
		
		let (y_min, y_max) = min_max_of_div_for_mult z x in
		Fcl_debug.call 'a' (fun s -> fprintf s "y_min=%d y_max=%d " y_min y_max);
		refine_min_max y ya y_min y_max;
		
		let (x_min, x_max) = min_max_of_div_for_mult z y in
		Fcl_debug.call 'a' (fun s -> fprintf s "x_min=%d x_max=%d\n" x_min x_max);
		refine_min_max x xa x_min x_max;
	        (* On rappelle update pour atteindre le point fixe *)
		if bounds <> compute_bounds () then loop () end in
    
    loop ();
	    
    Fcl_debug.call 'a' (fun s -> fprintf s "monome - after update : %a = %a * %a\n" Fd.fprint z Fd.fprint x Fd.fprint y)
  end;;

let monome x y =
  let e = Mult (Fd x, Fd y) in
  let (z_min, z_max) = min_max_of_expr e in
  let z = Fd.create (Fcl_domain.interval z_min z_max) in
  add_to_be_posted (z, Monome (x, y));
  z;;


let min_max_of_abs x =
  match Fd.value x with
    Val a -> let absa = abs a in (absa, absa)
  | Unk xa ->
      let a = Attr.min xa and b = Attr.max xa in
      if a >= 0 then (a, b) else if b <= 0 then (-b, -a) else
      (* x  cheval sur 0 *)
      let domx = Attr.dom xa in
      let a' = Fcl_domain.greatest_leq domx 0
      and b' = Fcl_domain.smallest_geq domx 0 in
      (min (-a') b', max (-a) b);;

class absolute z x = object
  inherit Fcl_cstr.t Fcl_cstr.normal

  method name = "absolute"

  method delay c =
    (* On pourrait suspendre on_refine sur x *)
    delay [Attr.on_min] z c; delay [Attr.on_max] z c;
    delay [Attr.on_min] x c; delay [Attr.on_max] x c

  method fprint c =
    Printf.fprintf c "%a = |%a|" Fd.fprint z Fd.fprint x; flush c

  method update =
    Fcl_debug.call 'a' (fun s -> fprintf s "absolute - before update : %a=|%a|\n" Fd.fprint z Fd.fprint x);
    
    begin match Fd.value x with
      Val a -> Fcl_stak.set solved true; Fd.unify z (abs a)
    | Unk xa ->
	if Attr.min xa >= 0 then begin
	  Fcl_stak.set solved true;
	  Fcl_cstr.post (new linear [(1, z); (-1, x)] Equal 0) end
	else if Attr.max xa <= 0 then begin
	  Fcl_stak.set solved true;
	  Fcl_cstr.post (new linear [(1, z); (1, x)] Equal 0) end
	else
	  match Fd.value z with
	    Val c ->
	      Fd.refine x
		(Fcl_domain.intersection
		   (Fcl_domain.create [-c; c]) (Attr.dom xa))
	  | Unk za ->
	      let (z_min, z_max) = min_max_of_abs x in
	      refine_min_max z za z_min z_max;

	      let x_min, x_max = Fd.min_max z in
	      assert (x_min >= 0);
	      Fd.refine x
		(Fcl_domain.remove_closed_inter
		   (-x_min+1) (x_min-1)
		   (Fcl_domain.remove_up x_max
		      (Fcl_domain.remove_low (-x_max) (Attr.dom xa)))) end;

    Fcl_debug.call 'a' (fun s -> fprintf s "absolute - after update : %a=|%a|\n" Fd.fprint z Fd.fprint x);
		
end

let absolute x =
  if Fd.min x >= 0 then Fd x
  else if Fd.max x <= 0 then Minus (Fd x) else
  let e = Abs (Fd x) in
  let (z_min, z_max) = min_max_of_expr e in
  let z = Fd.create (Fcl_domain.interval z_min z_max) in
  add_to_be_posted (z, Absolute x);
  Fd z;;


(* y = xr[a,b] / z[c,d] *)
(* a = min (x-r) , b = max (x-r) *)
let min_max_of_div_for_div a b z =
  (* Si 0 appartient au domaine de z, on ne peut rien dire sur y *)
  if Fd.member z 0 then
    (min_int, max_int)
  else
    let (c, d) = Fd.min_max z in
    (* Est-ce utile ? On ne rveille que sur min ou max *)
    (* z  cheval sur 0 *)
    if sign c * sign d < 0 then
      (* si z est instanci, c*d est positif *)
      match Fd.value z with
	Unk za ->
	  let domz = Attr.dom za in
	  let c' = Fcl_domain.greatest_leq domz 0
	  and d' = Fcl_domain.smallest_geq domz 0 in
	  (* z ne contient pas 0 *)
	  (min (a /+ d') (b /+ c'), max (a /- c') (b /- d'))
      |	_ -> Fcl_debug.internal_error "min_max_of_div_for_mult : z ground"
    (* z positif ou ngatif *)
    (* xr  cheval sur 0 *)
    else if sign a * sign b < 0 then
      if c > 0 then (a /+ c, b /- c) else (b /+ d, a /- d)
    (* ni xr ni z  cheval sur 0 *)
    else
      if a >= 0 then (* xr positif *)
	if c > 0 then (a /+ d, b /- c) else (b /+ d, a /- c) (* z positif ou ngatif *)
      else (* xr ngatif *)
	if c < 0 then (b /+ c, a /- d) else (a /+ c, b /- d) (* z ngatif ou positif *)

(* z=x/y, x=y*z+r *)
(* si x >= 0 alors r >= 0 sinon r <= 0 *)
let min_max_of_remainder x y =
  let r_abs_max =
    let min_y, max_y = Fd.min_max y in
    max (Pervasives.abs min_y) (Pervasives.abs max_y) - 1 in
  if Fd.min x >= 0 then (0, r_abs_max)
  else if Fd.max x <= 0 then ((0 - r_abs_max), 0)
  else ((0 - r_abs_max), r_abs_max);;

(* z = x / y *)
class division z x y =

  let div = Div (Fd x, Fd y)
  and yz_exp = Mult (Fd y, Fd z)
  and min_max_r () = min_max_of_remainder x y
  and compute_bounds () =
    (Fd.min z, Fd.max z, Fd.min x, Fd.max x, Fd.min y, Fd.max y) in
    
  
  object
  inherit Fcl_cstr.t Fcl_cstr.normal

  method name = "division"

  method delay c =
    delay [Attr.on_min] z c; delay [Attr.on_max] z c;
    delay [Attr.on_min] x c; delay [Attr.on_max] x c;
    delay [Attr.on_min] y c; delay [Attr.on_max] y c
      
  method fprint c =
    Printf.fprintf c "%a = %a / %a" Fd.fprint z Fd.fprint x Fd.fprint y;
    flush c

  (* On suppose que y ne contient pas 0 *)
  method update =
    Fcl_debug.call 'a' (fun s -> fprintf s "division - before update : %a = %a / %a\n" Fd.fprint z Fd.fprint x Fd.fprint y);

    let rec loop bounds =
      match Fd.value x, Fd.value y with
	Val a, Val b -> Fcl_stak.set solved true; Fd.unify z (a / b)
      | Val 0, _ -> Fcl_stak.set solved true; Fd.unify z 0
      | _x_val, _y_val ->
	  let (z_min, z_max) = min_max_of_expr div in
	  Fcl_debug.call 'a' (fun s -> fprintf s "z_min=%d z_max=%d " z_min z_max);
	  unifine z z_min z_max "division : _, _, Val c";
	  
	     (* x = y*z + r *)
	  let (yz_min, yz_max) = min_max_of_expr yz_exp
	  and (r_min, r_max) = min_max_r () in
	  let x_min = yz_min + r_min and x_max = yz_max + r_max in
	  Fcl_debug.call 'a' (fun s -> fprintf s "x_min=%d x_max=%d " x_min x_max);
	  unifine x x_min x_max "division : _, _, Val c";
	  
	     (* y = (x-r) / z *)
	  let (r_min, r_max) = min_max_r () in
	  let xr_min = Fd.min x - r_max and xr_max = Fd.max x - r_min in
	  let (y_min, y_max) = min_max_of_div_for_div xr_min xr_max z in
	  Fcl_debug.call 'a' (fun s -> fprintf s "y_min=%d y_max=%d\n" y_min y_max);
	  unifine y y_min y_max "division : _, _, Val c";

	  let new_bounds = compute_bounds () in
	  if bounds <> new_bounds then loop new_bounds in
    loop (compute_bounds ());
    
    Fcl_debug.call 'a' (fun s -> fprintf s "division - after update : %a = %a / %a\n" Fd.fprint z Fd.fprint x Fd.fprint y)
      
end

(* z = x / y *)
let division x y =
  begin match Fd.value y with
    Unk ya ->
      Fd.refine y (Fcl_domain.remove 0 (Attr.dom ya))
  | Val 0 -> raise Division_by_zero
  | _ -> () end;
  let e = Div (Fd x, Fd y) in
  let (z_min, z_max) = min_max_of_expr e in
  let z = Fd.create (Fcl_domain.interval z_min z_max) in
  add_to_be_posted (z, Division(x, y));
  z;;

(* y = (x-z) / (x/y) *)
let min_max_of_div_for_mod a b c d =
  (* Si 0 appartient au domaine de x/y, on ne peut rien dire sur y *)
  if c <= 0 && d >= 0 then
    (min_int, max_int)
  else (* x/y positif ou ngatif *)
    if sign a * sign b < 0 then (* xz  cheval sur 0 *)
      if c > 0 then (a /+ c, b /- c) else (b /+ d, a /- d)
    (* ni xz ni x/y  cheval sur 0 *)
    else if a >= 0 then (* xz positif *)
      if c > 0 then (a /+ d, b /- c) else (b /+ d, a /- c) (* x/y positif ou ngatif *)
    else (* xz ngatif *)
      if c < 0 then (b /+ c, a /- d) else (a /+ c, b /- d) (* x/y ngatif ou positif *)

(* z = x % y *)
class modulo z x y =

  let fdx = Fd x and fdy = Fd y and fdz = Fd z in
  let modxy = Mod (fdx, fdy) in
  let xy = Div (fdx, fdy) in
  let yxy = Mult (fdy, xy) in
  let x_exp = Plus (yxy, fdz) in
  let compute_bounds () =
    (Fd.min z, Fd.max z, Fd.min x, Fd.max x, Fd.min y, Fd.max y) in
  
  object
  inherit Fcl_cstr.t Fcl_cstr.normal

  method name = "modulo"

  method delay c =
    delay [Attr.on_min] z c; delay [Attr.on_max] z c;
    delay [Attr.on_min] x c; delay [Attr.on_max] x c;
    delay [Attr.on_min] y c; delay [Attr.on_max] y c
      
  method fprint c =
    Printf.fprintf c "%a = %a %% %a" Fd.fprint z Fd.fprint x Fd.fprint y;
    flush c

  (* On suppose que y ne contient pas 0 *)
  method update =
    Fcl_debug.call 'a' (fun s -> fprintf s "modulo - before update : %a = %a %% %a\n" Fd.fprint z Fd.fprint x Fd.fprint y);

    let rec loop bounds =
      match Fd.value x, Fd.value y with
	Val a, Val b -> Fcl_stak.set solved true; Fd.unify z (a mod b)
      | Val 0, _ -> Fcl_stak.set solved true; Fd.unify z 0
      | _x_val, _y_val ->
	  let (z_min, z_max) = min_max_of_expr modxy in
	  Fcl_debug.call 'a' (fun s -> fprintf s "z_min=%d z_max=%d " z_min z_max);
	  unifine z z_min z_max "modulo : _, _, Val c";
	  
	     (* x = y*(x/y) + z *)
	  let (x_min, x_max) = min_max_of_expr x_exp in
	  Fcl_debug.call 'a' (fun s -> fprintf s "x_min=%d x_max=%d " x_min x_max);
	  unifine x x_min x_max "modulo : _, _, Val c";
	  
	     (* y = (x-z) / (x/y) *)
	  let (xy_min, xy_max) = min_max_of_expr xy
	  and xz_min = Fd.min x - Fd.max z and xz_max = Fd.max x - Fd.min z in
	  let (y_min, y_max) =
	    min_max_of_div_for_mod xz_min xz_max xy_min xy_max in
	  Fcl_debug.call 'a' (fun s -> fprintf s "y_min=%d y_max=%d\n" y_min y_max);
	  unifine y y_min y_max "modulo : _, _, Val c";

	  let new_bounds = compute_bounds () in
	  if bounds <> new_bounds then loop new_bounds in
    loop (compute_bounds ());
    
    Fcl_debug.call 'a' (fun s -> fprintf s "modulo - after update : %a = %a %% %a\n" Fd.fprint z Fd.fprint x Fd.fprint y)

end

(* z = x % y *)
let modulo x y =
  begin match Fd.value y with
    Unk ya ->
      Fd.refine y (Fcl_domain.remove 0 (Attr.dom ya))
  | Val 0 -> raise Division_by_zero
  | _ -> () end;
  let e = Mod (Fd x, Fd y) in
  let (z_min, z_max) = min_max_of_expr e in
  let z = Fd.create (Fcl_domain.interval z_min z_max) in
  add_to_be_posted (z, Modulo (x, y));
  z;;


let expn_int x n =
  if n < 0 then raise (Invalid_argument "expn_int : negative exponent") else
  let rec loop = function
      0 -> 1
    | n ->
	let n2 = n / 2 in
	let xn2 = loop n2 in
	let xn = xn2 * xn2 in
	if n mod 2 = 0 then xn else xn * x in
  loop n

let min_max_of_expn x n =
  assert (n > 1);
  match Fd.value x with
    Unk xa ->
      let min_xa = Attr.min xa and max_xa = Attr.max xa in
      let min_xan = expn_int min_xa n and max_xan = expn_int max_xa n in
  (* L'exposant est pair *)
      if n mod 2 = 0 then
            (* Le domaine est positif *)
	if min_xa >= 0 then (min_xan, max_xan)
            (* Le domaine est ngatif *)
	else if max_xa <= 0 then (max_xan, min_xan)
            (* Le domaine contient des valeurs positives et ngatives *)
	else
	  let domx = Attr.dom xa in
	  let min_neg = Fcl_domain.greatest_leq domx 0
	  and min_pos = Fcl_domain.smallest_geq domx 0 in
	  let minn =
	    if -min_neg < min_pos then
	      expn_int min_neg n
	    else
	      expn_int min_pos n
	  and maxn = max min_xan max_xan in
	  (minn, maxn)
  (* L'exposant est impair *)
      else
	(min_xan, max_xan)
  | Val c -> let cn = expn_int c n in (cn, cn);;

let nth_root upper z n =
  let az = abs z and sz = sign z in
  (* Attention ! La racine d'un nombre ngatif renvoie nan ! *)
  let znth = truncate (float az ** (1. /. float n)) in
  if expn_int znth n = az then sz * znth else
  (* la calcule en float peut fournir z^(1/n)-e, donc avec le truncate -> -1 *)
  let znmore = znth + 1 in
  if expn_int znmore n = az then sz * znmore
  else if upper then
    if z >= 0 then znmore else (0 - znth)
  else if z >= 0 then znth else (0 - znmore)
  
let ( **/+) = nth_root true
let ( **/-) = nth_root false

(* Si n est pair, on ne renvoie que les bornes positives *)
let min_max_of_nth_root z n =
  (Fd.min z **/+ n, Fd.max z **/- n)
	

(* z = x^n *)
class expn z x n =

  let even = n mod 2 = 0

  and compute_bounds () = (Fd.min z, Fd.max z, Fd.min x, Fd.max x) in

  object
  inherit Fcl_cstr.t Fcl_cstr.normal

  method name = "expn"

  method delay c =
    delay [Attr.on_min] z c; delay [Attr.on_max] z c;
    delay [Attr.on_min] x c; delay [Attr.on_max] x c

  method fprint c =
    Printf.fprintf c "%a = %a ^ %d" Fd.fprint z Fd.fprint x n;
    flush c

  method update =
    Fcl_debug.call 'a' (fun s -> fprintf s "expn - before update : %a = %a ^ %d\n" Fd.fprint z Fd.fprint x n);

    let rec loop bounds =
      match Fd.value x, Fd.value z with
	Val a, _ -> Fcl_stak.set solved true; Fd.unify z (expn_int a n)
      | Unk xa, Val c ->
	  let root = truncate (float c ** (1. /. float n)) in
	  let rootn = expn_int root n in
	  if rootn = c then begin
	    Fcl_stak.set solved true;
	    if even then
	      Fd.refine x
		(Fcl_domain.intersection
		   (Fcl_domain.create [-root; root]) (Attr.dom xa))
	    else
	      Fd.subst x root end
	  else
	    Fcl_stak.fail "expn"
      | Unk xa, Unk za ->
	  let (z_min, z_max) = min_max_of_expn x n in
	  Fcl_debug.call 'a' (fun s -> fprintf s "expn - Unk xa, Unk za : z_min=%d z_max=%d " z_min z_max);	
	  Fd.refine z
	    (Fcl_domain.remove_up z_max
	       (Fcl_domain.remove_low z_min (Attr.dom za)));
	  let (x_min, x_max) = min_max_of_nth_root z n in
	  Fcl_debug.call 'a' (fun s -> if even then fprintf s "expn (even) - Unk xa, za : x_neg=[%d,%d] x_pos=[%d,%d]\n" (-x_max) (-x_min) x_min x_max else fprintf s "expn (odd) - Unk xa, Unk za : x_min=%d x_max=%d " x_min x_max);
	  if even then
	    (* [-x_max_pos, -x_min_pos] U [x_min_pos, x_max_pos] *)
	    Fd.refine x
	      (Fcl_domain.remove_closed_inter
		 (-x_min+1) (x_min-1)
		 (Fcl_domain.remove_up x_max
		    (Fcl_domain.remove_low (-x_max) (Attr.dom xa))))
	  else
	    (* [x_min, x_max] *)
	    Fd.refine x
	      (Fcl_domain.remove_up x_max
		 (Fcl_domain.remove_low x_min (Attr.dom xa)));

	  let new_bounds = compute_bounds () in
	  if bounds <> new_bounds then loop new_bounds in
    loop (compute_bounds ());
    
    Fcl_debug.call 'a' (fun s -> fprintf s "expn - after update : %a = %a ^ %d\n" Fd.fprint z Fd.fprint x n);

end

let expn x = function
    0 -> Fd.int 1
  | 1 -> x
  | n -> assert (n > 0);
      let (mini, maxi) = min_max_of_expn x n in
      let z = Fd.create (Fcl_domain.interval mini maxi) in
      add_to_be_posted (z, Expn (x, n));
      z;;


  (* sum xi = v ou sum xi <> v*)
  (* equal = true : equality, equal = false : inequality *)
class boolean_linear (terms : Fd.t list) v shared_min shared_max equal =
  let monotonic_propagate subst =
    List.fold_left
      (fun sum x ->
	match Fd.value x with
	  Val 0 -> sum
	| Val 1 -> 1 + sum
	| Unk attr ->
	    assert(Attr.dom attr = Fcl_domain.boolean);
	    Fd.subst x subst; subst + sum
	| _ -> Fcl_debug.internal_error "boolean_linear : non boolean variable")
      0
      terms in
	
  object
  inherit Fcl_cstr.t Fcl_cstr.normal

  method name = "boolean_linear"
  method delay c =
    List.iter (fun x -> delay [Attr.on_subst] x c) terms;
    delay [Attr.on_min] v c; delay [Attr.on_max] v c
  method fprint c =
    let shared_min = Fcl_stak.get shared_min
    and shared_max = Fcl_stak.get shared_max in
    Printf.fprintf c "%a %s[%d-%d] "
      Fd.fprint v (if equal then "=" else "<>") shared_min shared_max;
    List.iter (fun x -> Printf.fprintf c "+%a" Fd.fprint x) terms;
    flush c

  method update =
    let shared_min = Fcl_stak.get shared_min
    and shared_max = Fcl_stak.get shared_max in

    if equal then
    (* si on a atteint le maximum de v, on peut mettre tout les autres  zro*)
      if shared_min = Fd.max v then begin
      	Fcl_stak.set solved true;
      	if monotonic_propagate 0 > Fd.max v then
	  Fcl_stak.fail "boolean_linear: monotonic_propagate > max";
      	Fd.unify v shared_min
    (* Et vice versa *)
      end else if shared_max = Fd.min v then begin
      	Fcl_stak.set solved true;
       	if monotonic_propagate 1 < Fd.min v then
	  Fcl_stak.fail "boolean_linear: monotonic_propagate < min";
       	Fd.unify v shared_max
      end else if shared_min = shared_max then begin
 	Fd.unify v shared_min; Fcl_stak.set solved true
      end else if shared_max < Fd.min v || shared_min > Fd.max v then
	Fcl_stak.fail "boolean_linear (=)"
      else begin
	match Fd.value v with
	  Unk attr ->
	    Fd.refine v
	      (Fcl_domain.remove_up shared_max
		 (Fcl_domain.remove_low shared_min (Attr.dom attr)))
	| Val _ -> ()
      end
    else begin(* not equal *)
      if shared_min = shared_max then begin
	Fcl_stak.set solved true;
	match Fd.value v with
	  Val x -> 
	    if x = shared_min then
	      Fcl_stak.fail "boolean_linear (<>)"
	    (* else OK *)
	| Unk attr ->
	    Fd.refine v (Fcl_domain.remove shared_min (Attr.dom attr))
      end else if shared_min > Fd.max v || shared_max < Fd.min v then
	Fcl_stak.set solved true
    end

end;;
    
let set_cr op stakref = Fcl_stak.set stakref (op (Fcl_stak.get stakref) 1);;
let set_decr = set_cr (-);;
let set_incr = set_cr (+);;

class boolean_demon x shared_min shared_max = object
  inherit Fcl_cstr.t Fcl_cstr.immediate
  method name = "boolean_demon"
  method delay c = delay [Attr.on_subst] x c
  method fprint c = Fd.fprint c x

  method update =
    Fcl_stak.set solved true;
    match Fd.value x with
      Val 0 -> set_decr shared_max
    | Val 1 -> set_incr shared_min
    | _ -> Fcl_debug.internal_error "boolean_demon : variable is not ground or not boolean"
end;;
    
  
let is_boolean x =
  let min_x, max_x = Fd.min_max x in
  min_x = 0 && max_x = 1;;

(* On utilise boolean_linear et les dmons si le terme comprend
   plus de linear_bool_threshold boolens *)
let linear_bool_threshold = 5

let part_bools_others = List.partition (fun (_a, x) -> is_boolean x);;

let rec gather = function
    [] -> []
  | (a, x) :: bxs ->
      let rec les_autres axs = function
	  [] -> [(a, axs)]
	| ((a', x')::bxs) as bxs' ->
	    if a = a' then les_autres (x'::axs) bxs
	    else (a, axs) :: gather bxs' in
      les_autres [x] bxs ;;

(* flatten : ('a * 'b) list -> ('a * 'b list) list -> ('a * 'b) list *)
let rec flatten rest = function
    [] -> rest
  | (a, l) :: ls ->
      List.fold_right (fun x r -> (a,x)::r) l (flatten rest ls);;

(* liste trie suivant le coeff *)
let part_coef l =
  let ll = gather l in
  let (bools, too_short) = 
    List.partition
      (fun (_coeff, x) -> List.length x >= linear_bool_threshold) ll in
  (bools, flatten [] too_short );;

let boolean_linear bools sum_bi =
  let size = List.length bools in
  let shared_min = Fcl_stak.ref 0 and shared_max = Fcl_stak.ref size in
  List.iter (fun x -> Fcl_cstr.init (new boolean_demon x shared_min shared_max)) bools;
  new boolean_linear bools sum_bi shared_min shared_max true;;

let sum_boolean_linear bools =
  let size = List.length bools in
  let sum_bi = Fd.create (Fcl_domain.interval 0 size) in
  Fcl_cstr.post (boolean_linear bools sum_bi);
  sum_bi;;

let linear terms op d =
  (* On spare les variables boolennes des autres *)
  let (bools, others) = part_bools_others terms in
  (* On les trie suivant le coef *)
  let sorted_bools =
    List.sort (fun (a1, _x1) (a2, _x2) -> compare a1 a2) bools in
    (* boolss : (coef, size, bools) list; too_short est la liste des boolens
       qu'on a rejet parce-qu'il y en avait trop peu pour un meme coef *)
  let (boolss, too_short) = part_coef sorted_bools in
    (* Sum aj.sum_bij + others op d avec sum_bij = Sum bi *)
    (* On rcupre les (aj,sum_bij) *)
  let sums_bools =
      List.fold_left
      (fun a_sum_bis (a, bools) ->
	let sum_bi = sum_boolean_linear bools in (a, sum_bi) :: a_sum_bis)
      []
      boolss in
  
    (* On rifie la contrainte linaire globale : Sum aj.sum_bij + others op d *)
  new linear (sums_bools @ too_short @ others) op d;;
  

(* Attention : pas de rgle de simplification pour la division
   car 2*(5/2) <> (2*5)/2 *)
let rec mult e1 e2 = 
  match (e1, e2) with
    (Minus e1, e2) -> Minus (mult e1 e2)
  | (e1, Minus e2) -> Minus (mult e1 e2)
  | (Plus (e1,e2), e3) -> Plus (mult e1 e3, mult e2 e3)
  | (e3, Plus (e1,e2)) -> Plus (mult e1 e3, mult e2 e3)
  | (Int e1, Int e2) -> Int (e1*e2)
  | (Int 0, _) | (_, Int 0) -> Int 0
  | (Int 1, x) -> x
  | (x, Int 1) -> x
  | (e1, e2) -> Mult (e1, e2)

(* Disjunctive normal form *)
let rec simplifie = function
    Int _ as x -> x
  | (Fd v) as x ->
      begin
	match Fd.value v with
	  Val y -> Int y
	| _ -> x
      end
  | Plus (e1,e2) ->
      begin
      	match (simplifie e1, simplifie e2) with
	  (Int 0, x) -> x
	| (x, Int 0) -> x
	| (Int y, Int z) -> Int (y+z)
	| (s1, s2) -> Plus (s1, s2)
      end
  | Mult (e1, e2) ->
      let e1 = simplifie e1 and e2 = simplifie e2 in
      mult e1 e2
  | Minus e ->
      begin
	match simplifie e with
	  Int x -> Int (-x)
	| se -> Minus se
      end
  | Div (e1,e2) ->
      begin
	match (simplifie e1, simplifie e2) with
	  (Int 0, _) -> Int 0
	| (_, Int 0) -> raise Division_by_zero
	| (s1, s2) -> Div (s1, s2)
      end
  | Abs e ->
      begin
	match (simplifie e) with
	  Int x -> Int (Pervasives.abs x)
	| se -> Abs se
      end
  | Mod (e1,e2) ->
      begin
	match (simplifie e1, simplifie e2) with
	  (Int 0, _) -> Int 0
	| (_, Int 0) -> raise Division_by_zero
	| (e1, e2) ->
	    let mi1, ma1 = min_max_of_expr e1
	    and mi2 = min_of_expr e2 in
  	    if 0 <= mi1 && ma1 < mi2 then e1 else Mod (e1, e2)
      end

module Fd_for_H =
  struct
    type t = Fd.t
    let equal = Fd.equal
(* Attention ! Fd.name lve une exception sur les variables instancies *)
    let hash e = assert (Fd.is_var e); Hashtbl.hash (Fd.id e)
  end

module H = Hashtbl.Make(Fd_for_H)

  (* On construit la liste des variables prsentes dans la multiplication
  et de leur nombres d'occurences *)
let collect_mult vars =
  let hvars = H.create 11 in
  List.iter
    (fun v ->
      try incr (H.find hvars v)
      with Not_found -> H.add hvars v (ref 1))
    vars;
  let vs = ref [] in
  H.iter (fun v nb -> vs := (!nb, v) :: !vs) hvars;
  !vs;;

(* Transformation de PI(xi^ni) en z = x1n1*z1, x1n1=x1^n1 z1 = x2n2*z2,
  x2n2 = x2^n2 ... et retourne z *)
let overall vars =
  let vn_list = collect_mult vars in
  match vn_list with
    [] -> Int 1
  | (n1, v1) :: vns ->
      let whole =
	List.fold_left
	  (fun acc (n, v) -> monome acc (expn v n))
	  (expn v1 n1)
	  vns in
      Fd whole;;

(* On renvoie le coeff et la variable produit d'un monome *)
let mult_subst e =
  (* On fait remonter tous les coefficients entiers en les multipliant
     et on construit la liste des variables qui apparaissent dans e *)
  let rec coeff_first coeff vars = function
      Int x -> (x * coeff, vars)
    | Fd x -> begin
 	match Fd.value x with
	  Unk _ -> (coeff, x :: vars)
	| Val c -> (c * coeff, vars) end
    | Mult (e1, e2) ->
      	let (coeff_e1, vars1) = coeff_first coeff vars e1 in
      	coeff_first coeff_e1 vars1 e2
  (* On est cens n'avoir que des Mult, des Fd et des Int *)
    | _ -> Fcl_debug.internal_error "mult_subst : unexpected expression" in
  
  let (coeff_e, vars_e) = coeff_first 1 [] e in
  let v_e = overall vars_e in
  (coeff_e, v_e);;

(* Returns a new variable and an equality equation between this
variable and the given expression. Optimized for variables and integers *)
let bound_exp e eqs = 
  match e with
    Fd v -> (v, eqs)
  | Int i -> (Fd.int i, eqs)
  | _ ->
      let (min_e, max_e) = min_max_of_expr e in
      let v = Fd.create (Fcl_domain.interval min_e max_e) in
      (v, (v, e) :: eqs)

(* Removes Div's and Abs' and factorize them *)
let remove_div_abs e =
  let h = H_of_t.create 11 in
  let rec loop eqs = function
      (Int _ | Fd _) as e -> (e, eqs)
    | (Abs e) as e' ->
	begin
	  try
	    (H_of_t.find h e', eqs)
	  with
	    Not_found -> 
	      let (cleaned_e, eq_e) = loop eqs e in
	      let ve, eqs' = bound_exp cleaned_e eq_e in
	      let ve_abs = absolute ve in
	      H_of_t.add h e' ve_abs;
	      (ve_abs, eqs')
	end
    | (Div (e1, e2)) as e' ->
	begin
	  try
	    (H_of_t.find h e', eqs)
	  with
	    Not_found -> 
	      let (cleaned_e1, eq_e1) = loop eqs e1 in
	      let (cleaned_e2, eq_e2) = loop eq_e1 e2 in
	      let ve1, eqs1 = bound_exp cleaned_e1 eq_e2 in
  	      let ve2, eqs2 = bound_exp cleaned_e2 eqs1 in
	      let div12 = Fd (division ve1 ve2) in
	      H_of_t.add h e' div12;
	      (div12, eqs2)
	end
    | (Mod (e1, e2)) as e' ->
	begin
	  try
	    (H_of_t.find h e', eqs) (* Expression deja vue *)
	  with
	    Not_found -> 
	      let (cleaned_e1, eq_e1) = loop eqs e1 in
	      let (cleaned_e2, eq_e2) = loop eq_e1 e2 in
	      let ve1, eqs1 = bound_exp cleaned_e1 eq_e2 in
  	      let ve2, eqs2 = bound_exp cleaned_e2 eqs1 in
	      let mod12 = Fd (modulo ve1 ve2) in
	      H_of_t.add h e' mod12;
	      (mod12, eqs2)
	end
    | Mult (e1, e2) ->
	let (cleaned_e1, eq_e1) = loop eqs e1 in
	let (cleaned_e2, eq_e2) = loop eq_e1 e2 in
	(Mult (cleaned_e1, cleaned_e2), eq_e2)
    | Plus (e1, e2) ->
	let (cleaned_e1, eq_e1) = loop eqs e1 in
	let (cleaned_e2, eq_e2) = loop eq_e1 e2 in
	(Plus (cleaned_e1, cleaned_e2), eq_e2)
    | Minus e ->
	let (cleaned_e, eq_e) = loop eqs e in
	(Minus cleaned_e, eq_e) in
  loop [] e;;

let collect_vars sum =
    let vars = H.create 11 and constant = ref 0 in
    List.iter
      (function
	  (c, Fd x) when not (Fd.is_var x) ->
	    constant =+ ((Fd.min x)*c)
	| (c, Fd x) ->
	    (try
	      let (coeff,_x) = H.find vars x in
	      H.remove vars x;
	      H.add vars x (coeff+c, x)
	    with
	      Not_found -> H.add vars x (c, x))
  	| (c, Int x) -> constant =+ (x*c)
	| _ -> Fcl_debug.internal_error "collect_vars")
      sum;
    let l = ref [] in
    H.iter
      (fun _name (c,x) -> if c <> 0 then l := (c, x):: !l)
      vars;
    (!constant, !l);;

let big_sum e =
  let rec loop coeff l = function
      (Int _ | Fd _) as e -> (coeff, e) :: l
    (* simplifie assure qu'il n'y a que des Mult sous le Mult *)
    | Mult (_, _) as e ->
	let (coeff_e, ve) = mult_subst e in
	Fcl_debug.call 'a' (fun s -> fprintf s "mult : %d*%a\n" coeff_e fprint ve);
	(coeff_e * coeff, ve) :: l
    | Plus (e1, e2) -> loop coeff (loop coeff l e1) e2
    | Minus e -> loop (-coeff) l e
    | _ -> Fcl_debug.internal_error "big_sum : unexpected operator" in
  loop 1 [] e;;

let post_equation = function
    (z, Division (x, y)) -> Fcl_cstr.post (new division z x y)
  | (z, Monome (x, y)) -> Fcl_cstr.post (new monome z x y)
  | (z, Expn (x, n)) -> Fcl_cstr.post (new expn z x n)
  | (z, Absolute x) -> Fcl_cstr.post (new absolute z x)
  | (z, Modulo (x, y)) -> Fcl_cstr.post (new modulo z x y)

let rec search_and_replace v2 v1 = function
    [] -> raise Not_found
  | (z, rhs) as zrhs::eqs ->
      if Fd.equal z v2
      then (v1, rhs) :: eqs
      else zrhs :: search_and_replace v2 v1 eqs;;

let norm_expr equal_0 exp =
  let rec loop e =
    (* If an exception occurs during the iteration of post_equation
       on !to_be_posted (at the end of loop), !to_be_posted may not
       be empty in the next call to norm_expr. So we empty it. *)
    let linearize (v, e) =
      let (d, terms) = loop e in
      Fcl_cstr.post (linear ((-1, v) :: terms) Equal (-d)) in
    
    let simplified_e = simplifie e in
    let (cleaned_e, eqs) = remove_div_abs simplified_e in
    let coeff_vars = big_sum cleaned_e in
    let result = collect_vars coeff_vars in
  (* Intermediate constraints have been stored in 'to_be_posted' *)
    let parsed_result =
      (* Try to remove useless equation, e.g. when
	 x = x' and x' = y * z, x' is replaced by x *)
      match result with
      	(0, ([(1,v1);(-1,v2)])) when equal_0 -> begin
	  let (v2, v1) = if Fd.id v2 > Fd.id v1 then (v2, v1) else (v1, v2) in
	  try
	    to_be_posted := search_and_replace v2 v1 !to_be_posted;
	    Fcl_debug.call 'a' (fun s -> fprintf s "_%d replaced by _%d\n" (Fd.id v2) (Fd.id v1));
	    (0, [])
	  with
	    Not_found -> result
	end
	    (* An OR would be OK but buggy with ocamlc 3.01 *)
      | (0, ([(-1,v2);(1,v1)] )) when equal_0 -> begin
	  let (v2, v1) = if Fd.id v2 > Fd.id v1 then (v2, v1) else (v1, v2) in
	  try
	    to_be_posted := search_and_replace v2 v1 !to_be_posted;
	    Fcl_debug.call 'a' (fun s -> fprintf s "_%d replaced by _%d\n" (Fd.id v2) (Fd.id v1));
	    (0, [])
	  with
	    Not_found -> result
      end
      | _ -> result in
    List.iter linearize eqs;
    parsed_result
  in
  assert(!to_be_posted = []);
  try
    let r = loop exp in
    List.iter post_equation !to_be_posted;
    to_be_posted := [];
    r
  with
    exc ->
      to_be_posted := []; (* Restore a correct global state *)
      raise exc


  
      
      
let i2e x = Int x;;
let fd2e x = Fd x
let (+~) x y = Plus (x,y)
let sum tx = Array.fold_left (+~) (Int 0) tx
let sum_fd tx = sum (Array.map fd2e tx)
let ( *~) x y = Mult (x,y)
let (-~) x y = Plus (x,Minus y)
let (/~) x y = Div (x, y)
let (%~) x y = Mod (x, y)
let abs x = Abs x
let scalprod scals exps =
  if Array.length scals <> Array.length exps then
    raise (Invalid_argument "Arith.scalprod : arrays have not the same length");
  sum (Array.mapi (fun i ei -> ei *~ i2e scals.(i)) exps);;

let scalprod_fd ints vars = scalprod ints (Array.map fd2e vars);;

let ( **~) x n =
  if n < 0 then raise (Invalid_argument "**~ : negative exponent")
  else Fcl_misc.iter (fun r -> r *~ x) n (Int 1);;

let prod es = Array.fold_left ( *~) (Int 1) es

let prod_fd vs = prod (Array.map fd2e vs)

let do_it = function
    Leq0 (d, terms) -> linear terms LessThan (-d)
  | Neq0 (d, terms) -> linear terms Diff (-d)
  | Eq0 (d, terms) -> linear terms Equal (-d)


let (<=~) e1 e2 = do_it (Leq0 (norm_expr false (e1 -~ e2)))
let (=~) e1 e2 = do_it (Eq0 (norm_expr true (e1 -~ e2)))
let (<>~) e1 e2 = do_it (Neq0 (norm_expr false (e1 -~ e2)))
let (<~) e1 e2 = do_it (Leq0 (norm_expr false (i2e 1 +~ e1 -~ e2)))
let (>~) e1 e2 = e2 <~ e1
let (>=~) x y = (<=~) y x


let e2fd e =
  match simplifie e with
    Int x -> Fd.int x
  | Fd fd -> fd
  | e ->
      let (min_e, max_e) = min_max_of_expr e in
      let v = Fd.create (Fcl_domain.interval min_e max_e) in
      Fcl_cstr.post (fd2e v =~ e);
      v;;

let reify_bin op e1 e2 = fd2e (Fcl_reify.boolean (op e1 e2));;

let (=~~) = reify_bin (=~)
let (>=~~) = reify_bin (>=~)
let (<=~~) = reify_bin (<=~)
let (<~~) = reify_bin (<~)
let (>~~) = reify_bin (>~)
let (<>~~) = reify_bin (<>~)
