(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                 A Functional Constraint Library                     *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, CENA               *)
(*                                                                     *)
(* Copyright 2001 CENA. All rights reserved. This file is distributed  *)
(* under the terms of the GNU Lesser General Public License.           *)
(***********************************************************************)
(* $Id: fcl_cstr.ml,v 1.21 2001/05/22 20:05:17 brisset Exp $ *)

open Printf

exception DontKnow

type priority = int
let nb_priorities = 3

let immediate = 0
let normal = 1
let later = 2

class virtual t (p : priority) =
  let id = Fcl_misc.gen_int ()
  and woken = Fcl_stak.ref false in
object (self)
  val solved = Fcl_stak.ref false
  method id = id
  method is_solved = Fcl_stak.get solved
  method fprint c = fprintf c "%s" self#name
  method woken = woken
  method name = "anonymous"
  method virtual update : unit
  method init = ()
  method check = (Fcl_debug.fatal_error (self#name ^ ": check method undefined") : bool)
  method virtual delay : t -> unit
  method not = (Fcl_debug.fatal_error (self#name ^ ": not method undefined") : t)
  method priority = p
end

class create ?(priority = normal) ?fprint ?init ?check ?not ?name update delay = object
  inherit t priority as super
  method fprint =
    match fprint with None -> super#fprint | Some fprint -> fprint
  method name = match name with None -> super#name | Some name -> name
  method update = if update () then Fcl_stak.set solved true
  method init =
    match init with None -> super#init | Some init -> init ()
  method check =
    match check with None -> super#check | Some check -> check ()
  method delay = delay
  method not = match not with None -> super#not | Some not -> not ()
end

let fprint c ct = ct#fprint c

let create ?name ?fprint ?priority ?init ?check ?not update delay =
  new create ?priority ?fprint ?init ?check ?not ?name update delay

let queue = Array.create nb_priorities []
and already_in_wake = ref false
and next_priority = ref 0

let reset_queue () =
  for i = 0 to nb_priorities -1 do
    queue.(i) <- []
  done;
  next_priority := 0;
  already_in_wake := false;;

exception Wake_all

let wake_all () =
  if not !already_in_wake
  then begin
    already_in_wake := true;
    try
      while !next_priority < nb_priorities do
	match queue.(!next_priority) with
      	    [] -> incr next_priority
	  | (c : t)::cs ->
	      queue.(!next_priority) <- cs;
	      Fcl_debug.call 'c' (fun s -> fprintf s "%s#update\n" c#name);
	      c#update;
	      Fcl_debug.call 'c' (fun s -> fprintf s "%s#updated%s\n" c#name (if c#is_solved then "*" else ""));
	      Fcl_stak.unsafe_set c#woken false
      done;
      already_in_wake := false
   (* To avoid being in a state where already_in_wake = true after an
      uncaught exception during the while loop. *)
    with e -> (reset_queue (); raise e) end


let schedule_one_cstr =
  fun cstr ->
    if not (cstr#is_solved || Fcl_stak.get cstr#woken) then begin
      Fcl_debug.call 'c' (fun s -> fprintf s "wake: "; cstr#fprint s; fprintf s "\n");
      let p = cstr#priority in
      queue.(p) <- cstr :: queue.(p);
      next_priority := Fcl_misc.Operators.min !next_priority p;
      Fcl_stak.set cstr#woken true;
    end


(* Management of active contraints *)
module Store = struct
  let size_store = 1024
  let store = ref (Weak.create size_store)
  let next_free = ref 0
    
  let compress_or_extend () =
    let size = Weak.length !store in
    let rec look_for_free from to_ =
      if to_ < size then
      	match Weak.get !store to_ with
      	  None -> copy (max (to_ + 1) from) to_
      	| Some _ -> look_for_free from (to_ + 1)
      else size (* Full *)
    and copy from to_ =
      assert(from > to_);
      if from < size then
      	match Weak.get !store from with
      	  None -> copy (from + 1) to_
      	| some_c ->
	    Weak.set !store to_ some_c;
	    Weak.set !store from None;
	    look_for_free (from + 1) (to_ + 1)
      else to_ in
    next_free := look_for_free 0 0;
    if !next_free > size / 2 then begin
      let old_store = !store in
      store := Weak.create (size * 2);
      Weak.blit old_store 0 (!store) 0 size;
    end
      
  let add = fun c ->
    if not c#is_solved then begin
      let size = Weak.length !store in
      if !next_free >= size then begin
      	assert(!next_free = size);
      	compress_or_extend () (* Set next_free *)
      end;
      Weak.set !store !next_free (Some c);
      incr next_free;
      Fcl_stak.trail
      	(fun () ->
        (* le weak pointer de c a ete eventuellement supprime par le GC et
	   la compression *)
	  match Weak.get !store (!next_free - 1) with
	    Some c' when c'#id <> c#id -> ()
	  | _ -> decr next_free)
    end
	
  let active_store () =
    let rec loop active i = 
      if i < !next_free then
      	loop
	  (match Weak.get !store i with
	    None -> active
      	  | Some c ->
	      if c#is_solved then active else (c::active))
	  (i+1)
      else
      	active in
    loop [] 0
end

let active_store = Store.active_store
  

let post c =
  Fcl_debug.call 'c' (fun s -> fprintf s "post: "; c#fprint s; fprintf s "\n");
  let current_status = !already_in_wake in
  already_in_wake := true;
  c#init;
  already_in_wake := current_status;
  schedule_one_cstr c;
  wake_all ();
  if not c#is_solved then begin
    c#delay c;
    Store.add c;
  end;;

  (* post pour les dmons *)
let init c =
  c#init;
  c#delay c;;


class one = object
  inherit t immediate
  initializer Fcl_stak.set solved true
  method name = "one"
  method delay _ = ()
  method check = true
  method update = ()
  method not = new zero
end
and zero = object
  inherit t immediate
  method name = "zero"
  method delay _ = ()
  method check = false
  method update = Fcl_stak.fail "Cstr.zero"
  method not = new one
end

let one = new one
let zero = new zero

let id c = c#id
let name c = c#name

(* Un objet avec des contraintes qui lui sont attaches *)
type event = t list Fcl_stak.ref
let new_event () = Fcl_stak.ref []
let schedule event = List.iter schedule_one_cstr (Fcl_stak.get event)
let register event cstr =
  let current = Fcl_stak.get event in
  if not cstr#is_solved
  then Fcl_stak.set event (cstr :: current)
let registered = Fcl_stak.get

let delay events c =
  List.iter
    (fun event -> register event c)
    events

let conjunction cstrs =
  let update () = true
  and delay _ = ()
  and init () = List.iter (fun c -> post c) cstrs in
  create ~init:init ~name:"conjunction" update delay
