(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                 A Functional Constraint Library                     *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, CENA               *)
(*                                                                     *)
(* Copyright 2001 CENA. All rights reserved. This file is distributed  *)
(* under the terms of the GNU Lesser General Public License.           *)
(***********************************************************************)
(* $Id: fcl_var.mli,v 1.21 2001/06/19 13:18:30 barnier Exp $ *)

(* Module [Var]: Constrained, Attributed, Finite Domain Variables *)

(*** The Attribute of a Domain Variable *)
module Attr : sig
  type t
(* Type of attributes. *)
  val dom : t -> Fcl_domain.t
(* [dom a] returns the integer domain of an attribute. *)
  type event
(* Type of events (modifications on variables) on which to suspend
   a constraint. *)
  val on_refine : event
(* Event occuring when a variable is changed, i.e. its domain modified. *)
  val on_subst : event
(* Event occuring when a variable is instantiated. *)
  val on_min : event
  val on_max : event
(* Event occuring when the lower (resp. upper) bound of a variable
   decreases. *)
  val fprint : out_channel -> t -> unit
(* [fprint chan a] prints attribute [a] on channel [chan]. *)
  val size : t -> int
(* [size a] returns the number of integer values in the domain associated with
   [a] (i.e. [dom a]). *)
  val min : t -> int
  val max : t -> int
(* [min a] (resp. [max a]) returns the lower (resp. upper) bound of [a]. *)
  val values : t -> int list
(* [values a] returns the list of all integers in [dom a]. *)
  val iter : (int -> unit) -> t -> unit
(* [iter f a] iterates f on each integer in [dom a]. *)
  val member : t -> int -> bool
(* [member a n] tests if [n] belongs to [dom a]. *)
  val id : t -> int
(* [id a] returns a unique integer identifying the attribute [a]. *)
  val constraints_number : t -> int
(* [constraints_number a] returns the number of different constraints
   attached to [a]. *)
end

(*** A Domain Variable *)

type concrete_fd = Unk of Attr.t | Val of int
(* Type of the value of a Fd variable. *)

module Fd : sig
(* Finite domain variable *)
  type t
(* Type of Fd variable. *)

(*** Creation *)
  val create : ?name:string -> Fcl_domain.t -> t
(* [create ?name d] returns a new variable with domain [d]. If provided, [name]
   will be used by the pretty printer. *)
  val interval : ?name:string -> int -> int -> t
(* [interval ?name inf sup] returns a new variable with domain [[inf..sup]]. If
   provided, [name] will be used by the pretty printer.*)
  val array : ?name:string -> int -> int -> int -> t array
(* [array n inf sup] returns an array of [n] new variables with domain
   [[inf..sup]]. If provided, [name] (suffixed with the index of the element)
   will be used by the pretty printer. *)
  val int : int -> t
(* [int n] returns a new variable instantiated to integer value [n]. *)

(*** Access *)
  val is_var : t -> bool
(* [is_var v] returns [true] if [v] is not instantiated and [false]
   otherwise. *)
  val value : t -> concrete_fd
(* [value v] returns [Val n] if [v] is instantiated to [n], [Unk a] otherwise
   where [a] is the attribute of [v]. Should always be used in a matching:
   [match value v with Val n -> ... | Unk a -> ...]. *)
  val size : t -> int
(* [size v] returns the number of integer values in the domain of [v]
   ([1] if [v] is instantiated. *)
  val min : t -> int
(* [min v] returns the lower bound of [v]. *)
  val max : t -> int
(* [max v] returns the upper bound of [v]. *)
  val min_max : t -> int * int
(* [min_max v] returns both the lower and upper bounds of [v]. *)
  val int_value : t -> int
(* [int_value v] returns the value of [v] if it is instantiated and raises
   a [Failure] exception otherwise. *)
  val values : t -> int list
(* [values v] returns the list of all integers in the domain of [v]. If
   [v] is instantiated to [n], returns the singleton list containing [n]. *)
  val iter : (int -> unit) -> t -> unit
(* [iter f v] iterates f on each integer in the domain of [v]. *)
  val member : t -> int -> bool
(* [member v n] returns [true] if [n] belongs to the domain of [v] and
   [false] otherwise. *)
  val id : t -> int
(* [id v] returns a unique integer identifying the attribute associated
   with [v]. Must be called only on non ground variable, raise [Failure]
   otherwise. *)
  val name : t -> string
(* [name v] returns then attributed to variable [v] (the empty string if
   it was not provided while created). Must be called only on non ground
   variable, raise [Failure] otherwise. *)
  val compare : t -> t -> int
  (* Compares two variables. Values (bound variables) are smaller than
     unknowns (unbound variables). Unknowns are sorted according to
     their attribute [id]. *)
  val equal : t -> t -> bool
  (* Tests if two variables are equal with respect to [compare]. *)
  val fprint : out_channel -> t -> unit
(* [fprint chan v] prints variable [v] on channel [chan]. *)
  val fprint_array : out_channel -> t array -> unit
(* [fprint_array chan vs] prints array of variables [vs] on channel [chan]. *)

(*** Reduction *)
(*{*)
  val subst : t -> int -> unit
(* [subst v n] instantiates variable [v] with integer value [n]. Raises
   [Fcl_stak.Fail] in case of failure. Must be called only on unbound
   (not instantiated) variable, otherwise a [Failure] exception is raised. *)
(*}*)
  val unify : t -> int -> unit
(* [unify v n] instantiates variable [v] with integer value [n]. Raises
   [Fcl_stak.Fail] in case of failure. [unify] may be called either on unbound
   variables or on instantiated variables. *)
  val refine : t -> Fcl_domain.t -> unit
(* [refine v d] reduces the domain of [v] with domain [d]. [d] must be
   included in the domain of [v], otherwise the behaviour is
   unspecified (corrupted system or exception raised). *)
(*{*)
  val unify_cstr : t -> int -> Fcl_cstr.t
(*}*)
end

val delay : Attr.event list -> Fd.t -> Fcl_cstr.t -> unit
(* [delay event_list v c] suspends constraint [c] on all the events in
   [event_list] occurring on [v]. No effect on instantiated variables.
   To be used within new constraints "delay" functions.*)
